import numpy as np
import time

def load_GAP_instance(filename):
    try:
        with open(filename, "r") as file:
            # Reading the number of tasks and agents from the file
            tasks, agents = map(int, file.readline().split())

            # Reading the capacities of each agent from the file
            capacities = list(map(int, file.readline().split()))

            # Reading the costs matrix from the file
            costs = []
            for _ in range(tasks):
                costs.append(list(map(int, file.readline().split())))

        return tasks, agents, capacities, costs
    except IOError as e:
        print(f"Error opening or reading file: {e}")
        return None, None, None, None

def greedy_load_balancing_assignment(tasks, agents, capacities, costs):
    assigned_tasks = [-1] * tasks  # List to store tasks assigned to agents (-1 means unassigned)
    total_cost = 0
    remaining_capacity = capacities.copy()
    
    for task in range(tasks):
        best_agent = -1
        best_cost = float('inf')
        for agent in range(agents):
            if remaining_capacity[agent] > 0:  # If agent has capacity
                cost = costs[task][agent]
                if cost < best_cost or (cost == best_cost and remaining_capacity[agent] > remaining_capacity[best_agent]):
                    best_cost = cost
                    best_agent = agent
        if best_agent != -1:
            assigned_tasks[task] = best_agent
            remaining_capacity[best_agent] -= 1  # Reduce capacity of assigned agent
            total_cost += costs[task][best_agent]
        else:
            print(f"Task {task + 1} could not be assigned to any agent due to capacity constraints.")

    return assigned_tasks, total_cost

def calculate_total_cost(assigned_tasks, costs):
    total_cost = 0
    for task, agent in enumerate(assigned_tasks):
        if agent != -1:
            total_cost += costs[task][agent]
    return total_cost

def local_search(tasks, agents, capacities, costs, initial_solution):
    current_solution, current_cost = initial_solution
    best_solution = current_solution[:]
    best_cost = current_cost

    iterations_without_improvement = 0
    max_iterations = 1000

    while iterations_without_improvement < max_iterations:
        improved = False
        for task in range(tasks):
            current_agent = current_solution[task]
            for new_agent in range(agents):
                if new_agent != current_agent and capacities[new_agent] > 0:
                    # Try reassigning task to new_agent
                    new_solution = current_solution[:]
                    new_solution[task] = new_agent

                    # Check the new total cost
                    new_cost = calculate_total_cost(new_solution, costs)
                    if new_cost < best_cost:
                        best_solution = new_solution[:]
                        best_cost = new_cost
                        improved = True
                        iterations_without_improvement = 0
                        break

            if improved:
                break

        if not improved:
            iterations_without_improvement += 1

    return best_solution, best_cost

def main():
    filename = input("Enter the filename containing the GAP instance: ").strip()

    # Load GAP instance from file
    tasks, agents, capacities, costs = load_GAP_instance(filename)
    if tasks is None:
        return

    start_time = time.time()

    # Generate initial solution using a greedy heuristic
    initial_solution = greedy_load_balancing_assignment(tasks, agents, capacities, costs)

    # Apply local search to improve the initial solution
    best_solution, best_cost = local_search(tasks, agents, capacities, costs, initial_solution)

    # Print assigned tasks to agents
    print("Tasks assigned to agents after local search:")
    for agent in range(agents):
        tasks_assigned_to_agent = [str(task + 1) for task, assigned_agent in enumerate(best_solution) if assigned_agent == agent]
        print(f"Agent {agent + 1}: {', '.join(tasks_assigned_to_agent) if tasks_assigned_to_agent else 'None'}")

    end_time = time.time()
    execution_time = end_time - start_time

    print(f"Total cost: {best_cost}")
    print(f"Execution time: {execution_time:.3f} seconds")

if __name__ == "__main__":
    main()
