import numpy as np
import time

def load_GAP_instance(filename):
    try:
        with open(filename, "r") as file:
            tasks, agents = map(int, file.readline().split())
            capacities = list(map(int, file.readline().split()))
            costs = []
            for _ in range(tasks):
                costs.append(list(map(int, file.readline().split())))
        return tasks, agents, capacities, costs
    except IOError as e:
        print(f"Error opening or reading file: {e}")
        return None, None, None, None

def greedy_capacity_balancing(tasks, agents, capacities, costs):
    assigned_tasks = [-1] * tasks
    total_cost = 0
    remaining_capacity = capacities.copy()
    
    for task in range(tasks):
        best_agent = -1
        best_metric = float('inf')
        for agent in range(agents):
            if remaining_capacity[agent] > 0:
                metric = costs[task][agent] / (remaining_capacity[agent] + 1)
                if metric < best_metric:
                    best_metric = metric
                    best_agent = agent
        if best_agent != -1:
            assigned_tasks[task] = best_agent
            remaining_capacity[best_agent] -= 1
            total_cost += costs[task][best_agent]
        else:
            print(f"Task {task + 1} could not be assigned to any agent due to capacity constraints.")
    return assigned_tasks, total_cost

def calculate_total_cost(assigned_tasks, costs):
    total_cost = 0
    for task, agent in enumerate(assigned_tasks):
        if agent != -1:
            total_cost += costs[task][agent]
    return total_cost

def local_search(tasks, agents, capacities, costs, initial_assigned_tasks, initial_total_cost):
    current_assigned_tasks = initial_assigned_tasks.copy()
    current_total_cost = initial_total_cost
    best_assigned_tasks = initial_assigned_tasks.copy()
    best_total_cost = initial_total_cost
    improved = True

    while improved:
        improved = False
        for task in range(tasks):
            for agent in range(agents):
                if agent != current_assigned_tasks[task] and capacities[agent] > 0:
                    new_assigned_tasks = current_assigned_tasks.copy()
                    new_assigned_tasks[task] = agent
                    new_total_cost = calculate_total_cost(new_assigned_tasks, costs)
                    if new_total_cost < best_total_cost:
                        best_assigned_tasks = new_assigned_tasks.copy()
                        best_total_cost = new_total_cost
                        improved = True
        current_assigned_tasks = best_assigned_tasks.copy()
        current_total_cost = best_total_cost

    return best_assigned_tasks, best_total_cost

def main():
    filename = input("Enter the filename containing the GAP instance: ").strip()
    tasks, agents, capacities, costs = load_GAP_instance(filename)
    if tasks is None:
        return

    start_time = time.time()
    initial_assigned_tasks, initial_total_cost = greedy_capacity_balancing(tasks, agents, capacities, costs)
    best_assigned_tasks, best_total_cost = local_search(tasks, agents, capacities, costs, initial_assigned_tasks, initial_total_cost)

    print("Tasks assigned to agents:")
    for agent in range(agents):
        tasks_assigned_to_agent = [str(task + 1) for task, assigned_agent in enumerate(best_assigned_tasks) if assigned_agent == agent]
        print(f"Agent {agent + 1}: {', '.join(tasks_assigned_to_agent) if tasks_assigned_to_agent else 'None'}")

    end_time = time.time()
    execution_time = end_time - start_time

    print(f"Initial cost: {initial_total_cost}")
    print(f"Optimized cost: {best_total_cost}")
    print(f"Execution time: {execution_time:.2f} seconds")

if __name__ == "__main__":
    main()
