from scipy.spatial import distance_matrix
import numpy as np
import random
import time

# Function to generate random instances
def generate_instances(n, m, size):
    demand_points = np.random.rand(n, 2) * size  # Random coordinates in a sizeXsize grid
    facility_locations = np.random.rand(m, 2) * size  # Random coordinates in a sizeXsize grid
    dist_matrix = distance_matrix(demand_points, facility_locations)
    return demand_points, facility_locations, dist_matrix

# Calculate total distance
def calculate_total_distance(n, selected_facilities, dist_matrix):
    total_distance = 0
    for i in range(n):
        total_distance += min(dist_matrix[i, j] for j in selected_facilities)
    return total_distance

# Genetic Algorithm
def genetic_algorithm(n, m, p, dist_matrix, population_size=50, generations=100, mutation_rate=0.1):
    def initialize_population():
        return [random.sample(range(m), p) for _ in range(population_size)]
    
    def fitness(solution):
        return calculate_total_distance(n, solution, dist_matrix)
    
    def crossover(parent1, parent2):
        point = random.randint(1, p-1)
        child1 = parent1[:point] + [gene for gene in parent2 if gene not in parent1[:point]]
        child2 = parent2[:point] + [gene for gene in parent1 if gene not in parent2[:point]]
        return child1, child2
    
    def mutate(solution):
        if random.random() < mutation_rate:
            solution[random.randint(0, p-1)] = random.choice(range(m))
        return solution
    
    population = initialize_population()
    best_solution = min(population, key=fitness)
    best_fitness = fitness(best_solution)
    
    for _ in range(generations):
        new_population = []
        for _ in range(population_size // 2):
            parents = random.sample(population, 2)
            child1, child2 = crossover(parents[0], parents[1])
            new_population.extend([mutate(child1), mutate(child2)])
        population = new_population
        current_best = min(population, key=fitness)
        current_fitness = fitness(current_best)
        if current_fitness < best_fitness:
            best_solution, best_fitness = current_best, current_fitness
    
    return best_solution, best_fitness

# Local Search Heuristic
def local_search_specific(selected_facilities, dist_matrix, n, m, p, max_iterations=100):
    best_facilities = selected_facilities[:]
    best_total_distance = calculate_total_distance(n, best_facilities, dist_matrix)
    improved = True
    iteration = 0
    
    while improved and iteration < max_iterations:
        improved = False
        for i in best_facilities[:]:
            if i not in best_facilities:
                continue
            for j in set(range(m)) - set(best_facilities):
                new_selection = best_facilities[:]
                if i in new_selection:
                    new_selection.remove(i)
                new_selection.append(j)
                new_total_distance = calculate_total_distance(n, new_selection, dist_matrix)
                if new_total_distance < best_total_distance:
                    print(f"Improvement found at iteration {iteration}: {best_total_distance} -> {new_total_distance}")
                    best_facilities = new_selection[:]
                    best_total_distance = new_total_distance
                    improved = True
        iteration += 1
    
    return best_facilities, best_total_distance

if __name__ == "__main__":
    # Parameters
    n = int(input("Enter the number of demand points (n, max 500): "))
    m = int(input("Enter the number of potential facility locations (m, max 500): "))
    p = int(input("Enter the number of facilities to be selected (p, max 20): "))
    size = int(input("Enter the size of the grid to generate random coordinates: "))

    # Generate random instance
    demand_points, facility_locations, dist_matrix = generate_instances(n, m, size)

    # 1. Genetic Algorithm
    start_time = time.time()
    genetic_facilities, genetic_distance = genetic_algorithm(n, m, p, dist_matrix)
    genetic_time = time.time() - start_time

    # 2. Genetic Algorithm + Local Search
    start_time = time.time()
    genetic_ls_facilities, genetic_ls_distance = local_search_specific(genetic_facilities, dist_matrix, n, m, p)
    genetic_ls_time = time.time() - start_time

    # Results
    results = {
        "Genetic Algorithm": (genetic_distance, genetic_time),
        "Genetic Algorithm + Local Search": (genetic_ls_distance, genetic_ls_time)
    }
    
    for method, (distance, time_taken) in results.items():
        print(f"{method}: Total Distance = {distance}, Time = {time_taken:.4f} seconds")